module props
  use reader, only: grid1
  use tools_math
  implicit none

  public

  private :: promolecular_grid, index0, pri012, phi012, propwfn, promolecular, array_1d_std
  private :: grid1_interp
  private :: gascor, perdc

  integer :: iztype(120)
  type(grid1), allocatable :: grd(:)

contains

  subroutine init_rhogrid(m,nm)
    use reader
    use tools_io
    use param

    type(molecule), intent(in) :: m(nm)
    integer, intent(in) :: nm

    integer :: i, j, iz, nat, istat
    logical :: isat(120)

    isat = .false.
    do i = 1, nm
       do j = 1, m(i)%n
          iz = m(i)%z(j)
          isat(iz) = .true.
       end do
    end do

    nat = count(isat)
    allocate(grd(nat),stat=istat)
    if (istat /= 0) call error('nciplot','could not allocate memory for atomic density grids',faterr)

    nat = 0
    do i = 1, 120
       if (isat(i)) then
          nat = nat + 1
          iztype(i) = nat
          call readgrid(grd(nat),i)
       end if
    end do

  end subroutine init_rhogrid

  subroutine calcprops(x,m,nm,rho,rhom,grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r, anci)
    use reader
    use tools_io
    use param

    real*8, intent(in) :: x(3)
    integer, intent(in) :: nm
    type(molecule) :: m(nm)
    real*8, intent(out) :: rho, rhom(nm), grad(3), hess(3,3), dimgrad
    logical, intent(in) :: doelf
    integer, intent(in) :: ixc(2)
    real*8, intent(out) :: elf, exc
    real*8, intent(in)  :: acce_r
    type(averagenci) :: anci

    integer :: i
    real*8 :: hh(3,3), gg(3), rr, tp, tpp, rho53, grad2, ebose, df

    real*8, parameter :: fthird = -5d0/3d0
    real*8, parameter :: ctelf = 10d0*3d0**(-5d0/3d0)*pi**(-4d0/3d0)

    ! add by Pan Wu to solve doulbe counting density problem
    real*8 :: hh_lig(3,3), gg_lig(3), rr_lig, tpp_lig;
    integer :: tmp_count, std_count, index1, index2, index3, iframe
    integer :: nstep(3)
    real*8 :: p, q, r, tmp1
    real*8 :: wk1(3), wk2(3), heigs(3), hvecs(3,3)
    integer :: istat

    real*8, allocatable, dimension(:) :: array_dens ! , array_grad
    if(anci%do_std) then
        allocate(array_dens(1:anci%nframe),stat=istat)
        if (istat /= 0) call error('nciplot','could not allocate memory for temprary array',faterr)
!        allocate(array_grad(1:anci%nframe),stat=istat)
!        if (istat /= 0) call error('nciplot','could not allocate memory for temprary array',faterr)
    end if



    rho = 0d0
    rhom = 0d0
    grad = 0d0
    dimgrad = 0d0
    hess = 0d0
    tp = 0d0

    tmp_count = 0
    std_count = 0
    nstep(1) = anci%nstep(1)
    nstep(2) = anci%nstep(2)
    nstep(3) = anci%nstep(3)

    ! Need to remove the double counting of density for multiple frames, only the ligand
    ! if ligand in all files
    if(anci%ligand) then

       select case (m(anci%udat0)%ifile)
       case (ifile_xyz)
          ! promolecular densities
          call promolecular(x,m(anci%udat0),rr,gg,hh,acce_r)
          tpp = 0d0
       case (ifile_grd)
          call promolecular_grid(x,m(anci%udat0),rr,gg,hh)
          tpp = 0d0
       case (ifile_wfn)
          ! wfn densities
          call propwfn(x,m(anci%udat0),rr,gg,hh,tpp)
       case default
          call error('calcprops','unknown molecule type',faterr)
       end select
        
       ! weight rr, gg, hh
       rr_lig  = anci%fweight * rr
       gg_lig  = anci%fweight * gg
       hh_lig  = anci%fweight * hh
       tpp_lig = tpp

    end if
    

    do i = 1, nm
       select case (m(i)%ifile)
       case (ifile_xyz)
          ! promolecular densities
          call promolecular(x,m(i),rr,gg,hh,acce_r)
          tpp = 0d0
       case (ifile_grd)
          call promolecular_grid(x,m(i),rr,gg,hh)
          tpp = 0d0
       case (ifile_wfn)
          ! wfn densities
          call propwfn(x,m(i),rr,gg,hh,tpp)
       case default
          call error('calcprops','unknown molecule type',faterr)
       end select

       ! weight rr, gg, hh
       rr = anci%fweight * rr
       ! to make the dimension less gradient not have the scale from 0-1, the scale would be different
       gg = anci%fweight * gg
       hh = anci%fweight * hh
       ! DO NOT CONSIDER weight on tpp (BE CAREFUL!)

       ! real segments
       if(anci%lig_in_all .and. (i /= anci%udat0) .and. (anci%fweight > 0.99999)) then
           rr  =  rr  - rr_lig
           gg  =  gg  - gg_lig
           hh  =  hh  - hh_lig
           tpp =  tpp - tpp_lig

       ! real frames and with lig_in_all, need to pre-calculate reduced gradient
!       else if (anci%lig_in_all .and. (i /= anci%udat0) .and. (anci%fweight < 1.D0) .and. (anci%scheme .eq. 1)) then
!           grad2 = dot_product(gg,gg)
!           dimgrad = dimgrad + sqrt(grad2) / (const*rr**(4.D0/3.D0))           
!           tmp_count = tmp_count + 1
     
       ! no lig_in_all, which is trimmed, no change
       end if

       ! if estimate the fluctuation
       if(anci%do_std .and. i /= anci%udat0) then
           std_count = std_count + 1
!           grad2 = dot_product(gg,gg);
           array_dens(std_count) = rr
!           array_grad(std_count) = sqrt(grad2)
       end if

       ! Sum it up
       rho = rho + rr
       rhom(i) = rhom(i) + rr
       grad = grad + gg
       hess = hess + hh
       tp = tp + tpp

    end do


    ! for real frame, need to delete the ligand density again for consistent
    if(anci%ligand .and. (anci%fweight < 1.0D0) ) then
        rho = rho -rr_lig
        grad = grad - gg_lig
        hess = hess - hh_lig
        tp = tp - tpp_lig
    end if
  
    if(anci%do_std) then
        array_dens = array_dens * anci%nframe
!        array_grad = array_grad * anci%nframe
        call array_1d_std(array_dens, anci%nframe, anci%tmp_std_rho)
!        call array_1d_std(array_grad, anci%nframe, anci%tmp_std_grad)
        ! deallocate these two arrays
        deallocate(array_dens)
!        deallocate(array_grad)
    end if

    grad2 = dot_product(grad,grad)
    dimgrad = sqrt(grad2) / (const*rho**(4.D0/3.D0))

    exc = 0d0
    if (all(ixc /= 0)) call calcexc(ixc,exc,rho,grad,tp)

    elf = 0d0
    if (doelf) then
       rho53 = ctelf * rho**fthird
       grad2 = grad(1)*grad(1)+grad(2)*grad(2)+grad(3)*grad(3)
       ebose = 0.125d0 * grad2 / rho
       df = rho53 * (tp-ebose)
       elf = 1d0 / (1d0 + df*df)
    end if

  end subroutine calcprops

  !> Calculate the standard deviation of 1D array
  subroutine array_1d_std(array, nframe, final)
    integer, intent (in) :: nframe
    real*8, intent(in)   :: array(nframe)
    real*8, intent(out)  :: final

    integer :: i
    real*8 :: mean, var, total

    mean  = 0.0
    var   = 0.0
    total = 0.0

    do i = 1, nframe
        total = total + array(i);
        var = var + array(i)**2;
    end do

    mean = total/nframe;
    var = (var - total*mean)/(nframe-1);
    final = sqrt(var)

  end subroutine array_1d_std

  

  !> Calculate the density and gradient using atomic densities
  subroutine promolecular(x,m,rho,grad,hess,acce_r)
    use reader
    use tools_io
    use param

    real*8, intent(in) :: x(3)
    type(molecule), intent(in) :: m
    real*8, intent(out) :: rho, grad(3), hess(3,3)

    ! The C and ZETA data blocks are the coefficients and exponents
    ! for the atoms in the first three row of the periodic table, H-Ar.
    !    rho_atom = sum_i c_i * exp(-r / z_i) where i = 1..3
    ! The index runs over atomic numbers
    real*8, parameter :: c1(18) = (/&
       0.2815D0, 2.437D0, 11.84D0, 31.34D0, 67.82D0, 120.2D0, 190.9D0,&
       289.5D0,  406.3D0, 561.3D0, 760.8D0, 1016.D0, 1319.D0, 1658.D0,&
       2042.D0, 2501.D0, 3024.D0, 3625.D0/)
    real*8, parameter :: c2(18) = (/&
         0.D0,    0.D0, 0.06332D0, 0.3694D0, 0.8527D0, 1.172D0, 2.247D0,&
      2.879D0, 3.049D0,   6.984D0,  22.42D0,  37.17D0, 57.95D0, 87.16D0,&
      115.7D0, 158.0D0,   205.5D0,  260.0D0/)
    real*8, parameter :: c3(18) = (/&
       0.D0, 0.D0, 0.D0, 0.D0, 0.D0, 0.D0, 0.D0, 0.D0, 0.D0, 0.D0,&
       0.06358D0, 0.3331D0, 0.8878D0, 0.7888D0, 1.465D0, 2.170D0,&
       3.369D0, 5.211D0/)
    real*8, parameter :: zeta1(18) = (/&
       0.5288D0, 0.3379D0, 0.1912D0, 0.1390D0, 0.1059D0, 0.0884D0,&
       0.0767D0, 0.0669D0, 0.0608D0, 0.0549D0, 0.0496D0, 0.0449D0,&
       0.0411D0, 0.0382D0, 0.0358D0, 0.0335D0, 0.0315D0, 0.0296D0/)
    real*8, parameter :: zeta2(18) = (/&
       1.D0, 1.D0, 0.9992D0, 0.6945D0, 0.5300D0, 0.5480D0,&
       0.4532D0, 0.3974D0, 0.3994D0, 0.3447D0, 0.2511D0, 0.2150D0,&
       0.1874D0, 0.1654D0, 0.1509D0, 0.1369D0, 0.1259D0, 0.1168D0/)
    real*8, parameter :: zeta3(18) = (/&
       1.D0, 1.D0, 1.D0, 1.D0, 1.D0, 1.D0, 1.D0, 1.D0, 1.D0, 1.D0,&
       1.0236D0, 0.7753D0, 0.5962D0, 0.6995D0, 0.5851D0, 0.5149D0,&
       0.4974D0, 0.4412D0/)
    real*8, parameter :: czeta1(18) = c1 / zeta1
    real*8, parameter :: czeta2(18) = c2 / zeta2
    real*8, parameter :: czeta3(18) = c3 / zeta3
    real*8, parameter :: czzeta1(18) = czeta1 / zeta1
    real*8, parameter :: czzeta2(18) = czeta2 / zeta2
    real*8, parameter :: czzeta3(18) = czeta3 / zeta3

    integer :: i, iz, j, k
    real*8 :: exp1, exp2, exp3, r, fac0, fac1, fac2, xd(3), xu(3), xu2(3), r1, r2
    real*8 :: fac3

    ! add for screen function
    real*8 :: max_exp = 0.0
    real*8, intent(in) :: acce_r
!    integer :: count1, count2

    rho = 0d0
    grad = 0d0
    hess = 0d0

!    count1 = 0
!    count2 = 0

    do i = 1, m%n
       iz = m%z(i)
       if (iz < 1 .or. iz > atomic_zmax) &
          call error('promolecular','Atomic densities for high Z are not implemented (check atomic_zmax).',faterr)
       xd = x - m%x(:,i)
       r = sqrt(dot_product(xd,xd))

       ! add a screening function here
!       max_exp = maxval(/ zeta1(iz), zeta2(iz), zeta3(iz) /)
       if (zeta1(iz) > zeta2(iz)) then
           if(zeta1(iz) > zeta3(iz)) then
               max_exp = zeta1(iz)
           else 
               max_exp = zeta3(iz)
           end if
       else if( zeta2(iz) > zeta3(iz)) then
           max_exp = zeta2(iz)
       else 
           max_exp = zeta3(iz)
       end if
   
!write(*,*) zeta1(iz), zeta2(iz), zeta3(iz)
!write(*,*) max_exp, r
       if(acce_r > 0.0 .and. r > acce_r * max_exp) then
            cycle
       end if

       r1 = 1d0 / r
       r2 = r1 * r1
       xu = xd * r1
       xu2 = xu * xu

       exp1 = exp(-r/zeta1(iz))
       exp2 = exp(-r/zeta2(iz))
       exp3 = exp(-r/zeta3(iz))
       fac0 = c1(iz) * exp1 + c2(iz) * exp2 + c3(iz) * exp3
       fac1 = czeta1(iz)*exp1 + czeta2(iz)*exp2 + czeta3(iz)*exp3
       fac2 = czzeta1(iz)*exp1 + czzeta2(iz)*exp2 + czzeta3(iz)*exp3
       fac3 = fac2 + fac1 * r1

       rho = rho + fac0
       grad = grad - fac1 * xu
       ! hessian diagonal elements
       do j = 1, 3
          hess(j,j) = hess(j,j) + fac3 * xu2(j) - fac1 * r1
          do k = j+1, 3
             hess(j,k) = hess(j,k) + r2 * xd(j) * xd(k) * fac3
          end do
       end do
    end do
    do j = 1, 3
       do k = j+1, 3
          hess(k,j) = hess(j,k)
       end do
    end do
!write(*,*) count1, count2

  end subroutine promolecular

  subroutine propwfn(x,m,rho,g,hess,tp)
    use reader

    type(molecule), intent(in) :: m
    real*8, intent(in) :: x(3)
    real*8, intent(out) :: rho, g(3), hess(3,3), tp

    real*8 :: psi(m%nmo), f(10), store(m%npri,10)
    integer :: i, imo

    call phi012(x,m,store)

    rho = 0d0
    tp = 0d0
    g = 0d0
    hess = 0d0
    do imo = 1, m%nmo
       f = store(imo,1:10)
       ! Value of orbital
       psi(imo)=f(1)
       ! Density
       rho=rho+m%occ(imo)*f(1)*f(1)
       ! Gradient
       g = g + 2 * m%occ(imo) * f(1) * f(2:4)
       ! Diagonal Hessian
       do i = 1, 3
          hess(i,i) = hess(i,i) + 2 * m%occ(imo) * (f(1)*f(4+i)+f(1+i)**2)
       end do
       ! Off-diagonal Hessian
       hess(1,2) = hess(1,2) + 2 * m%occ(imo) * (f(1)*f(8)+f(2)*f(3))
       hess(1,3) = hess(1,3) + 2 * m%occ(imo) * (f(1)*f(9)+f(2)*f(4))
       hess(2,3) = hess(2,3) + 2 * m%occ(imo) * (f(1)*f(10)+f(3)*f(4))
       ! Kinetic Engery
       tp = tp + m%occ(imo) * (f(2)*f(2)+f(3)*f(3)+f(4)*f(4))
    end do
    tp = 0.5d0 * tp
    hess(2,1) = hess(1,2)
    hess(3,1) = hess(1,3)
    hess(3,2) = hess(2,3)

  end subroutine propwfn

  subroutine phi012(x,m,phi)
    use reader

    type(molecule), intent(in) :: m
    real*8, intent(in) :: x(3)
    real*8, intent(out) :: phi(m%npri,10)

    real*8 :: pri(m%npri,10)
    integer :: i, ipri, imo

    ! values of primitives
    call pri012(x,m,pri)

    ! sum
    phi = 0d0
    do i = 1, 10
       do ipri = 1, m%npri
          do imo = 1, m%nmo
             phi(imo,i) = phi(imo,i) + m%c(imo,ipri)*pri(ipri,i)
          end do
       end do
    end do

  end subroutine phi012

  subroutine pri012(x,m,pri)
    use reader
    use tools_io
    use param

    type(molecule), intent(in) :: m
    real*8, intent(in) :: x(3)
    real*8, intent(out) :: pri(m%npri,10)

    integer :: l(3), ipri, ic, it, i
    real*8 :: al, x0(3), x2, ex, xl(3,0:2)

    do ipri = 1, m%npri
       ic = m%icenter(ipri)
       it = m%itype(ipri)

       al = m%e(ipri)
       x0 = x - m%x(:,ic)
       x2 = dot_product(x0,x0)

       ! add a justification of whether cutoff is needed
!write(*,*) al*x2
       if(al*x2 > 12.0) then
          pri(ipri,:) = 0.0
          cycle
       end if

       ex = exp(-al * x2)

       call index0(it,l)
       do i = 1, 3
          if (l(i) == 0) then
             xl(i,0) = 1d0
             xl(i,1) = 0d0
             xl(i,2) = 0d0
          else if (l(i) == 1) then
             xl(i,0) = x0(i)
             xl(i,1) = 1d0
             xl(i,2) = 0d0
          else if (l(i) == 2) then
             xl(i,0) = x0(i) * x0(i)
             xl(i,1) = 2d0 * x0(i)
             xl(i,2) = 2d0
          else
             call error('pri012','power of L not supported',faterr)
          end if
       end do

       pri(ipri,1) = xl(1,0)*xl(2,0)*xl(3,0)*ex
       pri(ipri,2) = (xl(1,1)-2*al*x0(1)**(l(1)+1))*xl(2,0)*xl(3,0)*ex
       pri(ipri,3) = (xl(2,1)-2*al*x0(2)**(l(2)+1))*xl(1,0)*xl(3,0)*ex
       pri(ipri,4) = (xl(3,1)-2*al*x0(3)**(l(3)+1))*xl(1,0)*xl(2,0)*ex
       pri(ipri,5) = (xl(1,2)-2*al*(2*l(1)+1)*xl(1,0)&
          +4*al*al*x0(1)**(l(1)+2))*xl(2,0)*xl(3,0)*ex
       pri(ipri,6) = (xl(2,2)-2*al*(2*l(2)+1)*xl(2,0)&
          +4*al*al*x0(2)**(l(2)+2))*xl(3,0)*xl(1,0)*ex
       pri(ipri,7) = (xl(3,2)-2*al*(2*l(3)+1)*xl(3,0)&
          +4*al*al*x0(3)**(l(3)+2))*xl(1,0)*xl(2,0)*ex
       pri(ipri,8) = (xl(1,1)-2*al*x0(1)**(l(1)+1))*&
          (xl(2,1)-2*al*x0(2)**(l(2)+1))*xl(3,0)*ex
       pri(ipri,9) = (xl(1,1)-2*al*x0(1)**(l(1)+1))*&
          (xl(3,1)-2*al*x0(3)**(l(3)+1))*xl(2,0)*ex
       pri(ipri,10)= (xl(3,1)-2*al*x0(3)**(l(3)+1))*&
          (xl(2,1)-2*al*x0(2)**(l(2)+1))*xl(1,0)*ex
    end do

   end subroutine pri012

   subroutine index0(i,l)
     use tools_io
     use param

     integer, intent(in) :: i
     integer, intent(out) :: l(3)

     integer, parameter :: LI(3,35) = reshape((/&
        0,0,0, 1,0,0, 0,1,0, 0,0,1, 2,0,0, 0,2,0, 0,0,2, 1,1,0, 1,0,1, 0,1,1,&
        3,0,0, 0,3,0, 0,0,3, 2,1,0, 2,0,1, 0,2,1, 1,2,0, 1,0,2, 0,1,2, 1,1,1,& ! f primitives
        0,0,4, 0,1,3, 0,2,2, 0,3,1, 0,4,0, 1,0,3, 1,1,2, 1,2,1, 1,3,0, 2,0,2,& ! g primitives
        2,1,1, 2,2,0, 3,0,1, 3,1,0, 4,0,0/),shape(li))

     if (i < 1 .or. i > 35) call error('index0','type not allowed',faterr)
     l = li(:,i)

   end subroutine index0

  !> Calculate the density and gradient using atomic densities
  subroutine promolecular_grid(x,m,rho,grad,hess)
    use reader
    use tools_io
    use param

    real*8, intent(in) :: x(3)
    type(molecule), intent(in) :: m
    real*8, intent(out) :: rho, grad(3), hess(3,3)

    integer :: i, j, k, iz, ityp
    real*8 :: xd(3), r, r1, r2, f, fp, fpp, rfac, radd

    rho = 0d0
    grad = 0d0
    hess = 0d0
    do i = 1, m%n
       iz = m%z(i)
       ityp = iztype(iz)
       if (ityp == 0) call error('promolecular_grid','Atom type not initialized',faterr)
       if (.not.grd(ityp)%init) call error('promolecular_grid','Atom type not initialized',faterr)

       xd = x - m%x(:,i)
       r = sqrt(dot_product(xd,xd))
       r1 = 1d0 / r
       r2 = r1 * r1
       call grid1_interp(grd(ityp),r,f,fp,fpp)

       rho = rho + f
       grad = grad + fp * xd * r1
       rfac = (fpp - fp * r1)
       do j = 1, 3
          hess(j,j) = hess(j,j) + fp * r1 + rfac * r2 * xd(j) * xd(j)
          do k = j+1, 3
             radd = rfac * r2 * xd(j) * xd(k)
             hess(j,k) = hess(j,k) + radd
          end do
       end do
    end do
    do j = 1, 3
       do k = j+1, 3
          hess(k,j) = hess(j,k)
       end do
    end do

  end subroutine promolecular_grid

  !> Interpolate the radial grid g at distance r0, and obtain the value,
  !> first derivative and second derivative.
  subroutine grid1_interp(g,r0,f,fp,fpp)
    use reader, only: grid1

    type(grid1), intent(in) :: g !< The radial grid.
    real*8, intent(in) :: r0 !< Value of the radial coordinate.
    real*8, intent(out) :: f !< Interpolated value
    real*8, intent(out) :: fp !< Interpolated first derivative
    real*8, intent(out) :: fpp !< Interpolated second derivative

    integer :: ir, i, j, ii
    real*8 :: r, prod, rr(4), dr1(4), x1dr12(4,4)

    f = 0d0
    fp = 0d0
    fpp = 0d0

    if (.not.g%init) return
    if (r0 >= g%rmax) return

    ! careful with grid limits.
    if (r0 <= g%r(1)) then
       ir = 1
       r = g%r(1)
    else
       ir = 1 + floor(log(r0/g%a)/g%b)
       r = r0
    end if

    x1dr12 = 0d0
    do i = 1, 4
       ii = min(max(ir,2),g%ngrid-2) - 2 + i
       rr(i) = g%r(ii)
       dr1(i) = r - rr(i)
       do j = 1, i-1
          x1dr12(i,j) = 1d0 / (rr(i) - rr(j))
          x1dr12(j,i) = -x1dr12(i,j)
       end do
    end do

    ! interpolate, lagrange 3rd order, 4 nodes
    do i = 1, 4
       ii = min(max(ir,2),g%ngrid-2) - 2 + i
       prod = 1.d0
       do j = 1 ,4
          if (i == j) cycle
          prod = prod * dr1(j) * x1dr12(i,j)
       end do
       f = f + g%f(ii) * prod
       fp = fp + g%fp(ii) * prod
       fpp = fpp + g%fpp(ii) * prod
    end do

  end subroutine grid1_interp

  subroutine checkden()
    use reader
    use tools_io
    use param

    type(molecule) :: m

    integer, parameter :: mpts = 1000
    real*8, parameter :: rmax = 8d0

    integer :: i, iz
    real*8 :: r, x(3), rho1, rho2, grad1(3), grad2(3), hess1(3,3), hess2(3,3)

    integer, parameter :: lu = 10
    real*8 :: acce_r = 0.0

    ! define a hydrogen atom
    m%name = "h.dummy"
    m%ifile = ifile_xyz
    m%n = 1
    allocate(m%x(3,1))
    allocate(m%z(1))
    m%x = 0d0
    m%z = 1

    open(unit=lu,file="checkden.dat")
    do iz = 1, 18
       m%z = iz
       write (lu,'("# Z = ",I3)') iz
       do i = 1, mpts
          r = rmax * real(i,8) / real(mpts,8)
          x = (/0d0,0d0,r/)
          call promolecular(x,m,rho1,grad1,hess1,acce_r)
          call promolecular_grid(x,m,rho2,grad2,hess2)
          write (lu,*) i, r, rho1, rho2
       end do
       write (lu,*)
       write (lu,*)
    end do
    close(lu)

    stop 1

  end subroutine checkden

  subroutine calcexc(ixc,exc,rho,grad,tau)
    use tools_io
    use param
    implicit none

    integer, intent(in) :: ixc(2)
    real*8, intent(out) :: exc
    real*8, intent(in) :: rho, grad(3), tau

    real*8, parameter :: pol_2_13  = 2d0**(1d0/3d0)
    real*8, parameter :: pol_2_23  = 2d0**(2d0/3d0)
    real*8, parameter :: cslater   = 3d0/2d0 * (3d0/4d0/pi)**(1d0/3d0)
    real*8, parameter :: cslater_u = cslater / 2**(1d0/3d0)
    real*8, parameter :: b_b86     = 0.00336d0
    real*8, parameter :: b_b86_u   = b_b86 / 2**(1d0/3d0)
    real*8, parameter :: g_b86     = 0.00449d0
    real*8, parameter :: b_b88     = 0.0042d0
    real*8, parameter :: b_b88_u   = b_b88 / 2**(1d0/3d0)
    real*8 :: ex, ec, dum1, dum2, ucor0
    real*8 :: rho43, rdg, rdg2, asnh, rho2, grad2, tau2

    exc = 0d0
    ! Exchange
    select case (ixc(1))
    case (1)
       ! LDA exchange, Slater with alpha = 4/3, total rho
       ex = -cslater_u * rho**(4d0/3d0)
    case (2)
       ! PBE exchange (similar to B86), total rho
       rho43 = rho**(4d0/3d0)
       rdg2 = dot_product(grad,grad) / rho43**2 * pol_2_23
       ex = (-cslater_u -b_b86_u * rdg2 / (1d0 + g_b86 * rdg2)) * rho43
    case (3)
       ! Becke88 exchange, total rho
       rho43 = rho**(4d0/3d0)
       rdg = sqrt(dot_product(grad,grad)) / rho43 * pol_2_13
       rdg2 = rdg * rdg
       asnh = log(rdg+sqrt(rdg2+1d0))
       ex = (-cslater_u -b_b88_u * rdg2 / (1d0 + 6d0*b_b88*rdg*asnh)) * rho43
    case (-1)
       ex = 0d0
    case (99)
       ! test
       ! rs = (3d0 / (4d0*pi*rho))**(1d0/3d0)
       call pbex(rho,dot_product(grad,grad),1,ex,dum1,dum2)
       ex = -cslater_u * rho**(4d0/3d0) + ex
    case default
       call error('calcprops','unknown ixc option',faterr)
    end select

    ! Correlation
    select case (ixc(2))
    case (1)
       ! Perdew-Wang 88 parametrization of HEG eps_c
       call gascor(0.5d0*rho,0.5d0*rho,ec,dum1,dum2,ucor0)
    case (2)
       ! PBE correlation (similar to B86), total rho
       call perdc(0.5d0*rho,0.5d0*rho,sqrt(dot_product(grad,grad)),0d0,ec)
    case (3)
       ! Becke88 correlation, total rho
       rho2 = 0.5d0 * rho
       grad2 = 0.25d0 * dot_product(grad,grad)
       tau2 = 0.5d0 * tau
       call b88_corr(rho2,rho2,grad2,grad2,tau2,tau2,ec)
    case (-1)
       ec = 0d0
    case (99)
       ! test
       call gascor(0.5d0*rho,0.5d0*rho,ec,dum1,dum2,ucor0)
       call pbec(rho,dot_product(grad,grad),1,ucor0,dum1,dum2)
       ec = ec + ucor0
    case default
       call error('calcprops','unknown ixc option',faterr)
    end select
    exc = ex + ec

  end subroutine calcexc

  ! Erin Johnson private communication, Sep 14th 2011
  subroutine gascor(RHO1,RHO2,ECOR,VC1,VC2,UCOR)
    use param, only: pi, eps
    IMPLICIT REAL*8(A-H,O-Z)

    DATA PGAM,PFZZ/0.5198421D0,1.709921D0/
    THRD=1.D0/3.D0
    THRD4=4.D0*THRD
    !**** PW91 CORRELATION LSDA,
    PRS = (0.75D0/PI/(RHO1+RHO2+EPS))**THRD
    PZET = (RHO1-RHO2)/(RHO1+RHO2+EPS)/(1.D0+EPS)
    PF = ((1.D0+PZET)**THRD4+(1.D0-PZET)**THRD4-2.D0)/PGAM
    PA = 0.0310907D0
    PA1 = 0.21370D0
    PB1 = 7.5957D0
    PB2 = 3.5876D0
    PB3 = 1.6382D0
    PB4 = 0.49294D0
    PP = 1.00D0
    PP1 = PP + 1.D0
    PQ0 = -2.D0*PA*(1.D0+PA1*PRS)
    PRS12 = DSQRT(PRS)
    PRS32 = PRS12**3
    PRSP = PRS**PP
    PQ1 = 2.D0*PA*(PB1*PRS12+PB2*PRS+PB3*PRS32+PB4*PRS*PRSP)
    PQ2 = DLOG(1.D0+1.D0/PQ1)
    PEU = PQ0*PQ2
    PQ3 = PA*(PB1/PRS12+2.D0*PB2+3.D0*PB3*PRS12+2.D0*PB4*PP1*PRSP)
    PEURS = -2.D0*PA*PA1*PQ2-PQ0*PQ3/(PQ1**2+PQ1)
    PA = 0.01554535D0
    PA1 = 0.20548D0
    PB1 = 14.1189D0
    PB2 = 6.1977D0
    PB3 = 3.3662D0
    PB4 = 0.62517D0
    PP = 1.00D0
    PP1 = PP + 1.D0
    PQ0 = -2.D0*PA*(1.D0+PA1*PRS)
    PRSP = PRS**PP
    PQ1 = 2.D0*PA*(PB1*PRS12+PB2*PRS+PB3*PRS32+PB4*PRS*PRSP)
    PQ2 = DLOG(1.D0+1.D0/PQ1)
    PEP = PQ0*PQ2
    PQ3 = PA*(PB1/PRS12+2.D0*PB2+3.D0*PB3*PRS12+2.D0*PB4*PP1*PRSP)
    PEPRS = -2.D0*PA*PA1*PQ2-PQ0*PQ3/(PQ1**2+PQ1)
    PA = 0.0168869D0
    PA1 = 0.11125D0
    PB1 = 10.357D0
    PB2 = 3.6231D0
    PB3 = 0.88026D0
    PB4 = 0.49671D0
    PP = 1.00D0
    PP1 = PP + 1.D0
    PQ0 = -2.D0*PA*(1.D0+PA1*PRS)
    PRSP = PRS**PP
    PQ1 = 2.D0*PA*(PB1*PRS12+PB2*PRS+PB3*PRS32+PB4*PRS*PRSP)
    PQ2 = DLOG(1.D0+1.D0/PQ1)
    PALFM = PQ0*PQ2
    PQ3 = PA*(PB1/PRS12+2.D0*PB2+3.D0*PB3*PRS12+2.D0*PB4*PP1*PRSP)
    ALFRSM = -2.D0*PA*PA1*PQ2-PQ0*PQ3/(PQ1**2+PQ1)
    PALFC = -PALFM
    PZ4 = PZET**4
    PEC = PEU*(1.D0-PF*PZ4)+PEP*PF*PZ4-PALFM*PF*(1.D0-PZ4)/PFZZ
    ECOR = (RHO1+RHO2) * PEC
    PECRS = PEURS*(1.D0-PF*PZ4)+PEPRS*PF*PZ4-ALFRSM*PF*(1.D0-PZ4)/PFZZ
    PFZ = THRD4*((1.D0+PZET)**THRD-(1.D0-PZET)**THRD)/PGAM
    PECZET = 4.D0*(PZET**3)*PF*(PEP-PEU+PALFM/PFZZ)+PFZ*(PZ4*PEP-PZ4*PEU-(1.D0-PZ4)*PALFM/PFZZ)
    PCOMM = PEC -PRS*PECRS/3.D0-PZET*PECZET
    VC1 = PCOMM + PECZET
    VC2 = PCOMM - PECZET
    !**** END PW91 CORRELATION LSDA.
    PTC = -4.D0*PEC + 1.5D0*((1.D0+PZET)* VC1+(1.D0-PZET)* VC2)
    PTCOR = (RHO1+RHO2) * PTC
    UCOR =   ECOR-PTCOR
    !**** END PW91 CORRELATION KINETIC AND POTENTIAL ENERGIES.

  end subroutine gascor

  ! Erin Johnson private communication, Sep 14th 2011
  subroutine perdc(RHO1,RHO2,DROT,ZET,CPBE)
    use param, only: pi
    IMPLICIT REAL*8(A-H,O-Z)

    ROT=RHO1+RHO2
    CALL GASCOR(RHO1,RHO2,ECOR,VC1,VC2,UCOR)
    THRD=1.D0/3.D0
    THRD2=2.D0/3.D0
    FERMIK=(3.D0*PI*PI)**(1.D0/3.D0)
    PHI=0.5D0*((1.D0+ZET)**THRD2+(1.D0-ZET)**THRD2)
    PKF=FERMIK*ROT**THRD
    PKS=DSQRT(4.D0*PKF/PI)
    T=DROT/(2.D0*PHI*PKS*ROT)
    A=2.14612D0/(DEXP(-ECOR/ROT/0.031091D0/PHI**3)-1.D0)
    TSTUFF=(1.D0+A*T**2)/(1.D0+A*T**2+A**2*T**4)
    H=0.031091D0*PHI**3*DLOG(1.D0+2.14612D0*T**2*TSTUFF)
    CPBE=ECOR+ROT*H

  end subroutine perdc

  ! Erin Johnson private communication, Sep 14th 2011
  subroutine b88_corr(rho1,rho2,grad1,grad2,tau1,tau2,ec)
    use param, only: pi
    implicit none

    real*8, intent(in) :: rho1, rho2 !< rho, 2 spins
    real*8, intent(in) :: grad1, grad2 !< nablarho * nablarho, 2 spins
    real*8, intent(in) :: tau1, tau2 !< kinetic energy density, 2 spins
    real*8, intent(out) :: ec

    real*8, parameter :: thrd4 = 4d0/3d0
    real*8, parameter :: cex = 1.5D0*(0.75D0/pi)**(1.D0/3.D0)

    real*8 :: drhosq, drho, weizs1, weizs2, rho43, dlss, xl, asnh
    real*8 :: x88, rfg1, rfg2, z11, z22, z12, c8811, c8812, c8822
    real*8 :: dsigs1, dsigs2

    ! SPIN 1:
    weizs1 = 0.25d0 * grad1 / rho1
    drhosq = 4.D0 * rho1 * weizs1
    drho = sqrt(drhosq)
    rho43 = rho1**thrd4
    dlss = drho / rho43
    xl = -cex * rho43
    asnh = log(dlss+sqrt(dlss**2+1.D0))
    x88 = xl - 0.0042D0*drho*dlss/(1.D0+6.D0*0.0042D0*dlss*asnh)
    rfg1 = -0.5D0*rho1/x88
    z11 = 0.88D0*2.D0*rfg1
    dsigs1 = 2d0 * tau1 - weizs1
    c8811 = -0.01D0*rho1*dsigs1*z11**4*(1.D0-2.D0/z11*log(1.D0+0.5D0*z11))

    ! SPIN 2:
    weizs2 = 0.25d0 * grad2 / rho2
    drhosq = 4.D0 * rho2 * weizs2
    drho = sqrt(drhosq)
    rho43 = rho2**thrd4
    dlss = drho/rho43
    xl = -cex*rho43
    asnh = log(dlss+dsqrt(dlss**2+1.d0))
    x88 = xl - 0.0042d0*drho*dlss/(1.d0+6.d0*0.0042d0*dlss*asnh)
    rfg2 = -0.5d0*rho2/x88
    z22 = 0.88d0*2.d0*rfg2
    dsigs2 = 2d0 * tau2 - weizs2
    c8822 = -0.01d0*rho2*dsigs2*z22**4*(1.d0-2.d0/z22*log(1.d0+0.5d0*z22))

    ! opp spins:
    z12=0.63d0*(rfg1+rfg2)
    c8812=-0.8d0*rho1*rho2*z12**2*(1.d0-log(1.d0+z12)/z12)

    ec = c8811 + c8822 + c8812

  end subroutine b88_corr

end module props
