program nciplot
  use param
  use tools_io
  use tools_math
  use reader
  use props
  !ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
  !                NNN    NNN    CCCCCCCCCC   IIIIIIII
  !                NNN NN NNN   CCCCCCCCCCC     IIII
  !                NN  NN NNN   CCC              II
  !                NN   N NNN   CCC              II
  !                NN    NNNN   CCC              II
  !                NN     NNN   CCC              II
  !                NN     NNN   CCCCCCCCCCC     IIII
  !                NN     NNN    CCCCCCCCCC   IIIIIIII
  !
  ! Version: 2.0
  !
  !ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
  implicit none

  integer, parameter :: mfiles = 10000 !< only applies to rhom().

  character*(mline) :: argv(2), oname
  integer :: argc, nfiles, ifile, idx, istat, ntotal
  integer :: i, j, k, nn0, nnf, lp, n1
  real*8  :: tmp1
  character*(mline) :: filein, line, oline, word, wx, wc
  logical :: ok
  real*8 :: rdum
  ! the molecular info
  type(molecule), allocatable :: m(:)
  ! logical units
  integer :: lugc, ludc, luvmd, lusvmd, ludat, lusdat, luelf, luxc, lugsc, ludsc, lursc
  ! cubes
  real*8, allocatable, dimension(:,:,:) :: crho, cgrad, celf, cxc
  ! ligand, intermolecular keywords
  logical :: ligand, inter, intra, cpi
  real*8 :: rthres
  integer :: udat0
  ! radius and cube keywords
  logical :: autor
  real*8 :: x(3), xinit(3), xmax(3), xinc(3)
  integer :: nstep(3)
  ! noutput keyword
  integer :: noutput
  ! cutoffs
  real*8 :: rhocut, dimcut
  ! cutplot
  real*8 :: rhoplot, dimplot
  ! CPI cutoff
  real*8 :: cutgradcp, cutrhocp
  ! discarding rho parameter
  real*8 :: rhoparam
  ! properties of rho
  real*8 :: rho, grad(3), dimgrad, grad2, hess(3,3), elf, exc
  real*8 :: rhom(mfiles)
  ! eispack
  real*8 :: wk1(3), wk2(3), heigs(3), hvecs(3,3)
  ! elf
  logical :: doelf
  ! xc
  integer :: ixc(2)
  ! averagenci structure, add by Pan Wu
  type(averagenci) :: anci

  integer :: tmp = 0
  logical :: self_dens = .false.
  real*8  :: acce_r = 0.0

  character*(mline) :: tmp_word
  logical :: switcher

  ! initialize anci
!  anci%scheme  = 2     ! default as first average, then calculate
  anci%nframe  = 1
  anci%fweight = 1.0
  anci%ligand  = .false.
  anci%lig_in_all  = .false.

  anci%do_std = .false.

  ! initialize
  call param_init()

  ! I/O units, process arguments
  call getargs(argc,argv)
  if (argc >= 1) then
     open (uin,file=argv(1),status='old')
     if (argc == 2) then
        open(uout,file=argv(2),status='unknown')
     endif
  endif

  ! header
  call header()
  call tictac(' # Start')

  ! read files
  read (uin,*) nfiles
  if (nfiles > mfiles) call error('nciplot','too many files, increase mfiles',faterr)
  allocate(m(nfiles),stat=istat)
  if (istat /= 0) call error('nciplot','could not allocate memory for molecules',faterr)
  ntotal = 0
  do ifile = 1, nfiles
     read (uin,'(a)') filein
     filein = trim(adjustl(filein))
     inquire(file=filein,exist=ok)
     if(.not.ok) &
        call error('nciplot','requested file does not exist: '//trim(filein),faterr)
     m(ifile) = readfile(filein)
     if (ifile == 1) oname = filein(1:index(filein,'.',.true.)-1)
     ntotal = ntotal + m(ifile)%n
  enddo

  ! read density grids (props)
  call init_rhogrid(m,nfiles)

  ! by default, use density grids for heavier atoms
  do i = 1, nfiles
     if (m(i)%ifile == ifile_xyz .and. any(m(i)%z > atomic_zmax)) then
        m(i)%ifile = ifile_grd
        do j = 1, m(i)%n
           if (m(i)%z(j) > atomic_zmax .and..not.grd(iztype(m(i)%z(j)))%init) then 
              call error('nciplot','Some atomic density grids for heavy atoms are needed but not initialized',faterr)
           end if
        end do
     end if
  end do

  ! default values
  rhocut = 0.2d0
  xinc = 0.1d0
  if (any(m(:)%ifile == ifile_wfn)) then
     dimcut = 2.0d0
     dimplot=0.5d0
     rhoplot=0.05d0
  else
     dimcut = 1.0d0
     dimplot=0.3d0
     rhoplot=0.07d0
  end if
  rhoparam = 0.95d0
  noutput = 3
  udat0 = 1
  autor = .true.
  ligand = .false.
  inter = .false.
  rthres = 2.d0
  doelf = .false.
  ixc = 0
  ! read the rest of (optional) keywords
  xinit = m(1)%x(:,1)
  xmax = m(1)%x(:,1)
  do i = 1, nfiles
     do j = 1, m(i)%n
        xinit = min(xinit,m(i)%x(:,j))
        xmax = max(xmax,m(i)%x(:,j))
     end do
  end do
  do while (.true.)
     read (uin,'(a)',end=11) line
     line = trim(adjustl(line))
     oline = line
     call upper(line)
     if (line(1:1) == "#") cycle ! skip comments
     if (len(trim(line)) < 1) cycle ! skip blank lines

     idx = index(line,' ')
     word = line(1:idx-1)
     line = line(idx:)
     oline = oline(idx:)
     select case(trim(word))
     case ("LIGAND")
        ligand = .true.
        anci%ligand = .true.
        inter = .true.
        read (line,*) udat0, rthres, tmp
        anci%udat0 = udat0
        rthres = rthres / bohrtoa
        if(tmp) then
            anci%lig_in_all = .true.
        else
            anci%lig_in_all = .false.
        end if
     case ("INTERMOLECULAR")
        inter = .true.
     case ("FRAME")
        ! Get the frame # and ready for average frame, add by Pan, 2011-12
        read (line,*) anci%nframe
        anci%fweight = 1.0 / ( anci%nframe )
     case ("CPI")
        cpi = .true.
        read (line, *) read(line,*) cutgradcp,cutrhocp

     ! Estimate standard deviation of density and gradient
     case ("A-NCI-STD")
        anci%do_std = .true.

     case ("ACCE_R")
        read (line,*) acce_r
     case ("RADIUS")
        autor=.false.
        read(line,*) x, rdum
        xinit = (x - rdum) / bohrtoa
        xmax = (x + rdum) / bohrtoa
     case ("ONAME")
        read(oline,'(a)') oname
        oname = trim(adjustl(oname))
        oname = oname(1:index(oname,' '))
     case ("OUTPUT")
        read(line,*) noutput
     case ("CUBE")
        autor = .false.
        read(line,*) xinit, xmax
        xinit = xinit / bohrtoa
        xmax = xmax / bohrtoa
     case ("ATCUBE")
        autor = .false.
        xinit = 1d40
        xmax = -1d40
        do while (.true.)
           read (uin,'(a)') line
           line = trim(adjustl(line))
           call upper(line)
           if (line(1:1) == "#") cycle ! skip comments
           if (len(trim(line)) < 1) cycle ! skip blank lines
           lp = 1
           idx = index(line,' ')
           word = line(1:idx-1)
           if (trim(word) /= "END" .and. trim(word) /= "ENDATCUBE") then
              ok = isinteger(ifile,line,lp)
              if (.not.ok) call error('nciplot','bad atcube syntax',faterr)
              if (ifile < 1.or. ifile > nfiles) call error('nciplot','atcube: wrong file number',faterr)
              ok = isinteger(n1,line,lp)
              do while (ok)
                 if (n1<1.or.n1>m(ifile)%n) call error('nciplot','atcube: wrong atom number',faterr)
                 xinit = min(xinit,m(ifile)%x(:,n1))
                 xmax = max(xmax,m(ifile)%x(:,n1))
                 ok = isinteger(n1,line,lp)
              end do
           else
              exit
           end if
        end do
        xinit = xinit - rthres
        xmax = xmax + rthres
     case ("INCREMENTS")
        read(line,*) xinc
        xinc = xinc / bohrtoa
     case ("CUTOFFS")
        read(line,*) rhocut, dimcut
     case ("CUTPLOT")
        read(line,*) rhoplot, dimplot
     case ("RHOCUT2")
        read(line,*) rhoparam
     case ("ELF")
        doelf = .true.
     case ("EXC")
        ! exchange
        read (line,*) wx, wc
        if (trim(wx) == 'S' .or. trim(wx) == 'SLATER' .or. &
            trim(wx) == 'LDA') then
           ixc(1) = 1
        else if (trim(wx) == 'PBE') then
           ixc(1) = 2
        else if (trim(wx) == 'B88') then
           ixc(1) = 3
        else if (trim(wx) == '-') then
           ixc(1) = -1
        elseif (trim(wx) == 'TEST') then
           ixc(1) = 99
        else
           call error('nciplot','Unknown exchange functional',faterr)
        end if
        ! correlation
        write (*,*) trim(wc)
        if (trim(wc) == 'PW' .or. trim(wc) == 'LDA' ) then
           ixc(2) = 1
        else if (trim(wc) == 'PBE') then
           ixc(2) = 2
        else if (trim(wc) == 'B88') then
           ixc(2) = 3
        else if (trim(wc) == '-') then
           ixc(2) = -1
        elseif (trim(wc) == 'TEST') then
           ixc(2) = 99
        else
           call error('nciplot','Unknown correlation functional',faterr)
        end if
     case ("DGRID")
        do i = 1, nfiles
           if (m(i)%ifile == ifile_xyz) m(i)%ifile = ifile_grd
        end do
     !case ("CHECKDEN")
     !   call checkden()
     case default
        call error('nciplot','Don''t know what to do with '//trim(word)//' keyword',faterr)
     end select
  enddo
11 continue

  ! set grid limits and npts
  if (autor) then
     if (ligand) then
        xinit = m(udat0)%x(:,1)
        xmax = m(udat0)%x(:,1)
        do j = 1, m(udat0)%n
           xinit = min(xinit,m(udat0)%x(:,j))
           xmax = max(xmax,m(udat0)%x(:,j))
        end do
     end if
     xinit = xinit - rthres
     xmax = xmax + rthres
  end if
  nstep = ceiling((xmax - xinit) / xinc)

  ! punch info
  write(uout,124)
  if (inter) write(uout,126) 
  if (ligand) write(uout,123) trim(m(udat0)%name)
  write(uout,120)
  write(uout,110) 'RHO  THRESHOLD   (au):', rhocut
  write(uout,110) 'RDG  THRESHOLD   (au):', dimcut
  if (inter) write(uout,110) 'DISCARDING RHO PARAM :',rhoparam
  if (ligand) write(uout,110) 'RADIAL THRESHOLD  (A):',rthres*bohrtoa
  write(uout, 110)                            'ACCE_PARM    (number):', acce_r
  if (anci%nframe > 1) write(uout, '(A, I5)') 'MULTI-FRAME    (snap):', anci%nframe
  write (uout,*)
  write(uout,121) xinit, xmax, xinc, nstep

  ! open output files
  if (doelf) then
     luelf = 9
     open(luelf,file=trim(oname)//"-elf.cube")
  end if
  if (all(ixc /= 0)) then
     luxc = 8
     open(luxc,file=trim(oname)//"-xc.cube")
  end if
  if (noutput >= 2) then
     lugc = 11
     ludc = 12
     luvmd = 13
     open(lugc,file=trim(oname)//"-grad.cube")
     open(ludc,file=trim(oname)//"-dens.cube")
     open(luvmd,file=trim(oname)//".vmd")
  else
     lugc = -1
     ludc = -1
     luvmd = -1
  end if
  if (noutput == 1 .or. noutput == 3) then
     ludat = 14
     open(ludat,file=trim(oname)//".dat")
  else
     ludat = -1
  end if
  write(uout,122) trim(oname)//"-grad.cube",&
     trim(oname)//"-dens.cube",&
     trim(oname)//"-elf.cube",&
     trim(oname)//"-xc.cube",&
     trim(oname)//".vmd",&
     trim(oname)//".dat"
     
  ! write cube headers
  if (lugc > 0) call write_cube_header(lugc,'grad_cube','3d plot, reduced density gradient')
  if (ludc > 0) call write_cube_header(ludc,'dens_cube','3d plot, density')
  if (doelf) call write_cube_header(luelf,'elf_cube','3d plot, electron localisation function')
  if (all(ixc /= 0)) call write_cube_header(luxc,'xc_cube','3d plot, xc energy density')

  ! allocate memory for density and gradient
  if (ludc > 0) then
     allocate(crho(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat)
     if (istat /= 0) call error('nciplot','could not allocate memory for density cube',faterr)
  end if
  if (lugc > 0) then
     allocate(cgrad(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat)
     if (istat /= 0) call error('nciplot','could not allocate memory for grad',faterr)
  end if
  if (doelf) then
     allocate(celf(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat)
     if (istat /= 0) call error('nciplot','could not allocate memory for elf',faterr)
  end if
  if (all(ixc /= 0)) then
     allocate(cxc(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat)
     if (istat /= 0) call error('nciplot','could not allocate memory for xc',faterr)
  end if
  ! allocate memory for A-NCI density and gradient, if needed
  ! nframe, z, y, x, density
!  if (anci%do_cpt) then
!     allocate(anci%srho(1:anci%nframe+1, 1:nstep(3), 1:nstep(2), 1:3), stat=istat);
!     if (istat /= 0) call error('nciplot','could not allocate memory for A-NCI density',faterr)
!     allocate(anci%sdimgrad(1:anci%nframe+1, 1:nstep(3), 1:nstep(2), 1:3), stat=istat);
!     if (istat /= 0) call error('nciplot','could not allocate memory for A-NCI gradient',faterr)
!     allocate(anci%isinter(1:anci%nframe+1, 1:nstep(3), 1:nstep(2), 1:3), stat=istat);
!     if (istat /= 0) call error('nciplot','could not allocate memory for A-NCI Is-Inter',faterr)
!     allocate(anci%criticalpt(1:6, 10000), stat=istat);
!     if (istat /= 0) call error('nciplot','could not allocate memory for A-NCI critical points',faterr)
!   end if
 
   ! allocate space to save standard deviation of density and gradient at one point
   if(anci%do_std) then
       lugsc = 16
       ludsc = 17
       lursc = 18
       open(ludsc,file=trim(oname)//"-dens-std.cube")
!       open(lugsc,file=trim(oname)//"-grad-std.cube")
!       open(lursc,file=trim(oname)//"-rdg-std.cube")
       call write_cube_header(ludsc, 'dens_std_cube', '3d plot, density standard deviation')
!       call write_cube_header(lugsc, 'grad_std_cube', '3d plot, gradient standard deviation')
!       call write_cube_header(lursc, 'rdg_std_cube', '3d plot, averaged Reduced Gradient Density standard deviation')
       allocate(anci%std_rho(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat);
       if (istat /= 0) call error('nciplot','could not allocate memory for density  deviation cube',faterr);
!       allocate(anci%std_grad(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat);
!       if (istat /= 0) call error('nciplot','could not allocate memory for gradient deviation cube',faterr);
!       allocate(anci%std_rdg(0:nstep(3)-1,0:nstep(2)-1,0:nstep(1)-1),stat=istat);
!       if (istat /= 0) call error('nciplot','could not allocate memory for reduced density gradient deviation cube',faterr);
       lusdat = 19
       open(lusdat,file=trim(oname)//"-dens-std.dat")
       lusvmd = 20
       open(lusvmd,file=trim(oname)//"-std.vmd")
   else 
       lugsc = -1
       ludsc = -1
       lursc = -1
       lusvmd = -1
   end if
   
   if (anci%cpi) then
       lucpi = 21
       open(lucpi,file=trim(oname)//"-cpi.dat")
   endif

   anci%nstep = nstep
   anci%xinc  = xinc
   anci%xinit  = xinit
   anci%rhoparam = rhoparam


  
  write(uout, "('Status: [0%   10%   20%   30%   40%   50%   60%   70%   80%   90%   100%]')") 
  write(uout, "('Process: ',$)")

  tmp1 = real(nstep(1))/20.0 - 1.0
  ! calculate density and rdg
  !$omp parallel do private (x,rho,grad,hess,heigs,hvecs,wk1,wk2,istat,grad2,&
  !$omp dimgrad,intra,rhom,elf) schedule(dynamic)
  switcher = .false.
  do i = 0, nstep(1)-1
     do j = 0, nstep(2)-1
        do k = 0, nstep(3)-1
           x = xinit + (/i,j,k/) * xinc

           svec = xinit + (/i,j,k/)*xinc
           
           if(real(i) >= tmp1) then
               write(uout, "('>>>',$)")
               tmp1 = tmp1 + real(nstep(1))/20.0
           end if

           ! add a pre-selection, whether this grid is closer to ANY ligand molecule or not 
           ! if not, directly assign density as zero, rdg as 100.0D, std as zero
           if( ligand .eq. .true. ) then
                   call screengrid(switcher)
                   if(switcher .eq. .false.) then
                       crho(k,j,i) = 0.0
                       cgrad(k,j,i) = 100d0
                       if (anci%do_std) then
                           anci%std_rho(k,j,i)  = 0.0
                           ! anci%std_grad(k,j,i) = 0.0
                           ! anci%std_rdg(k,j,i)  = 0.0
                       end if
                       cycle;
                   end if
           end if

           ! calculate properties at x
           call calcprops(x,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           call rs(3,3,hess,heigs,0,hvecs,wk1,wk2,istat)


           grad2 = dot_product(grad,grad)
           dimgrad = sqrt(grad2) / (const*rho**(4.D0/3.D0))           

           ! rescale the dimgrad
!           dimgrad = (anci%fweight ** 0.333333) * dimgrad

           if (cpi) then
           if (rho.lt.cutrhocp.and.dimgrad.lt.cutgradcp) then

 2210    format ('rho grad ',2(f9.5,1x))
 2211    format ('out i j k rho ',3(i3,1x),2(f9.5,1x))
 2212    format ('in i, j, k ,rho ',3(i3,1x),2(f9.5,1x))

           call calcprops(x,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           call rs(3,3,hess,heigs,0,hvecs,wk1,wk2,istat)
           
           store=dimgrad
           storer=rho
           storeh=heigs(2)
           testvec(1)=svec(1) - xinc(1)
           testvec(2)=svec(2)
           testvec(3)=svec(3)

           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           testvec(1)=svec(1) + xinc(1)
           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           testvec(2)=svec(2) - xinc(2)
           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           testvec(2)=svec(2) + xinc(2)
           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           testvec(3)=svec(3) - xinc(3)
           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           testvec(3)=svec(3) + xinc(3)
           call calcprops(testvec,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           if (dimgrad.lt.store) then
               goto 10
           endif

           i4=i4+1
           ncp=nnuc3+i4
           xcp(ncp,1)=svec(1)
           xcp(ncp,2)=svec(2)
           xcp(ncp,3)=svec(3)
           rhocp(ncp)=sign(storer,storeh)
           dimcp(ncp)=store
           isusp(ncp)=0




10       CONTINUE

           call calcprops(x,m,nfiles,rho,rhom(1:nfiles),grad,dimgrad,hess,doelf,elf,ixc,exc,acce_r,anci)
           call rs(3,3,hess,heigs,0,hvecs,wk1,wk2,istat)

           
           intra = .false.
           if ((rho < rhocut) .and. (dimgrad < dimcut)) then

              ! for different frame
              if(anci%nframe .eq. 1) then
                  ! default as .false.
                  self_dens = (any(rhom(1:nfiles) >= rho*rhoparam))
              else 
                  self_dens = .false.
                  if (  (rhom(udat0) / (anci%fweight * rho)) >= rhoparam .or. (rhom(udat0) / (anci%fweight * rho)) <= (1.0-rhoparam) ) then
                      self_dens = .true.
                  end if
              end if
                    
              ! check if the interaction is localized on one of the fragments
              intra = ( (ligand.or.inter) .and. self_dens)

              ! write the dat file
              if(ludat > 0 .and..not.intra) then
                 !$omp critical (IO)
                 write(ludat,'(1p,E18.10,E18.10)') sign(rho,heigs(2)), dimgrad
                 !$omp end critical (IO)
                 if (anci%do_std) then
                     write(lusdat,'(1p,E18.10,E18.10)') sign(anci%tmp_std_rho / rho, heigs(2)), dimgrad
                 end if
              end if
           end if ! rhocut/dimcut
           
           ! write the cube files
           if (rho > rhoplot .or. intra) dimgrad = 100d0
           !$omp critical (cubewrite)
           if (lugc > 0) cgrad(k,j,i) = dimgrad
           if (ludc > 0) crho(k,j,i) = sign(rho,heigs(2))*100.D0
           if (doelf) celf(k,j,i) = elf
           if (all(ixc /= 0)) cxc(k,j,i) = exc
           !$omp end critical (cubewrite)
           ! save A-NCI standard deviation
           if (anci%do_std) then
!               anci%std_rho(k,j,i)  = sign(anci%tmp_std_rho / rho, heigs(2))   ! change to the relative density std.
               anci%std_rho(k,j,i)  = anci%tmp_std_rho / rho   ! change to the relative density std.
!               anci%std_grad(k,j,i) = anci%tmp_std_grad
               ! X = C*A*B^(-4/3)
               ! DELTA(X) = C*SQRT[ (DELTA(A) * B^(-4/3))^2 + ((4/3)*B^(-7/3)*A*DELTA(B))^2 ]
               ! A is the sqrt(grad), B is the rho
!write(*,*) rho, sqrt(grad2), anci%tmp_std_rho
!write(*,*) ( 1.0 / const), (anci%tmp_std_grad * rho**(-4.0/3.0) )**2.0, ( 4.0/3.0 * sqrt(grad2) * rho**(-7.0/3.0) * anci%tmp_std_rho)**2.0
!               anci%std_rdg(k,j,i)  = ( 1.0 / const) * sqrt( (anci%tmp_std_grad * rho**(-4.0/3.0) )**2.0 + ( 4.0/3.0 * sqrt(grad2) * rho**(-7.0/3.0) * anci%tmp_std_rho)**2.0 )
           end if
        end do
     end do
  end do
  !$omp end parallel do
  write(uout, "(' Done!')")

  ! write cubes
  if (ludc > 0) call write_cube_body(ludc,nstep,crho)
  if (lugc > 0) call write_cube_body(lugc,nstep,cgrad)
  if (doelf) call write_cube_body(luelf,nstep,celf)
  if (all(ixc /= 0)) call write_cube_body(luxc,nstep,cxc)
  if(anci%do_std) then
      call write_cube_body(ludsc,nstep,anci%std_rho)
!      call write_cube_body(lugsc,nstep,anci%std_grad)
!      call write_cube_body(lursc,nstep,anci%std_rdg)
  end if

  ! deallocate grids and close files
  if (allocated(crho)) deallocate(crho)
  if (allocated(cgrad)) deallocate(cgrad)
  if (allocated(celf)) deallocate(celf)
  if (allocated(cxc)) deallocate(cxc)
  if (ludat > 0) close(ludat)
  if (lusdat> 0) close(lusdat)


!  ! deallocate anci grids
!   if(allocated(anci%srho)) deallocate(anci%srho)
!   if(allocated(anci%sdimgrad)) deallocate(anci%sdimgrad)
!   if(allocated(anci%isinter)) deallocate(anci%isinter)
!   if(allocated(anci%criticalpt)) deallocate(anci%criticalpt)
   if(allocated(anci%std_rho)) deallocate(anci%std_rho)
!   if(allocated(anci%std_grad)) deallocate(anci%std_grad)
!   if(allocated(anci%std_rdg)) deallocate(anci%std_rdg)

  ! write vmd script
  if (ligand) then
     nn0 = sum(m(1:udat0-1)%n) + 1
     nnf = sum(m(1:udat0-1)%n) + m(udat0)%n
  else
     nn0 = 1
     nnf = ntotal
  end if
  if (luvmd > 0) then
     write (luvmd,114) trim(oname)//"-dens.cube"
     write (luvmd,115) trim(oname)//"-grad.cube"
     if (doelf) write (luvmd,115) trim(oname)//"-elf.cube"
     if (all(ixc /= 0)) write (luvmd,115) trim(oname)//"-xc.cube"
     write (luvmd,116) nn0-1,nnf-1,dimplot,2,2,2,-rhoplot*100D0,rhoplot*100D0,2,2
     close(luvmd)
  end if

  if (lusvmd > 0) then
     write (lusvmd,114) trim(oname)//"-dens-std.cube"
     write (lusvmd,115) trim(oname)//"-grad.cube"
     if (doelf) write (luvmd,115) trim(oname)//"-elf.cube"
     if (all(ixc /= 0)) write (luvmd,115) trim(oname)//"-xc.cube"
     write (lusvmd,116) nn0-1,nnf-1,dimplot,2,2,2,-rhoplot*100D0,rhoplot*100D0,2,2
     close(lusvmd)
  end if

  ! end
  call tictac('End')

  ! close files
  if (uin /= stdin) close(uin)
  if (uout /= stdout) close(uout)

110 format (A,F5.2)
114 format ('#!/usr/local/bin/vmd',/,&
     '# VMD script written by save_state $Revision: 1.2 $',/,&
     '# VMD version: 1.8.6            ',/,&
     'set viewplist            ',/,&
     'set fixedlist            ',/,&
     '# Display settings            ',/,&
     'display projection   Orthographic            ',/,&
     'display nearclip set 0.000000            ',/,&
     '# load new molecule         ',/,&
     'mol new ',a,' type cube first 0 last -1 step 1 filebonds 1 autobonds 1 waitfor all')
115 format ('mol addfile ',a,' type cube first 0 last -1 step 1 filebonds 1 autobonds 1 waitfor all')
116 format ('#',/,&
     '# representation of the atoms',/,&
     'mol delrep 0 top',/,&
     'mol representation Lines 1.00000',/,&
     'mol color Name',/,&
     'mol selection {all}',/,&
     'mol material Opaque',/,&
     'mol addrep top',/,&
     'mol representation CPK 1.000000 0.300000 118.000000 131.000000',/,&
     'mol color Name',/,&
     'mol selection {index ', i5, ' to ', i5, ' }',/,&
     'mol material Opaque',/,&
     'mol addrep top',/,&
     '#',/,&
     '# add representation of the surface',/,&
     'mol representation Isosurface ',f7.5,' 1 0 0 1 1',/,&
     'mol color Volume 0',/,&
     'mol selection {all}',/,&
     'mol material Opaque',/,&
     'mol addrep top',/,&
     'mol selupdate ',i1,' top 0',/,&
     'mol colupdate ',i1,' top 0',/,&
     'mol scaleminmax top ',i1,' ', f7.4, f7.4,/,&
     'mol smoothrep top ', i1,' 0',/,&
     'mol drawframes top ',i1,' {now}',/,&
     'color scale method BGR',/,&
     'set colorcmds {{color Name {C} gray}}',/,&
     '#some more',/)
120 format(/'-----------------------------------------------------'/&
            '      Calculation details:'/&
            '-----------------------------------------------------')
121 format(/,'-----------------------------------------------------'/&
             '      Operating grid and increments:'/&
             '-----------------------------------------------------'/&
             ' x0,y0,z0  = ',f8.4,' ',f8.4,' ',f8.4/&
             ' x1,y1,z1  = ',f8.4,' ',f8.4,' ',f8.4/&
             ' ix,iy,iz  = ',f5.2,'   ',f5.2,'   ',f5.2/&
             ' nx,ny,nz  = ',i4,'    ', i4,'    ', i4/)
122 format('-----------------------------------------------------'/&
           '      Writting output in the following units:'/&
           '-----------------------------------------------------'/&
           ' Reduced Density Gradient,RDG  = ',a,/&
           ' Sign(lambda2)xDensity,LS      = ',a,/&
           ' ELF cube file                 = ',a,/&
           ' XC energy density cube file   = ',a,/&
           ' LS x RDG                      = ',a,/&
           ' VMD script                    = ',a,/&
           '-----------------------------------------------------',/)
123 format('      Using ',a40,' as LIGAND')
124 format('-----------------------------------------------------'/&
           '      INPUT INFORMATION:'/&
           '-----------------------------------------------------')
126 format(/'      MIND YOU'/&
            '      ONLY ANALYZING INTERMOLECULAR INTERACTIONS     '/)

contains
  subroutine write_cube_header(lu,l1,l2)

    integer, intent(in) :: lu
    character*(*), intent(in) :: l1, l2

    integer :: i, j
    integer :: tmp_ntotal, tmp_count

    ! Maximum atom in CUBE is 99999
    tmp_count = 0
    tmp_ntotal = ntotal;

    if(tmp_ntotal > 99999) then
        tmp_ntotal = 99999
    end if

    write(lu,*) trim(l1)
    write(lu,*) trim(l2)
    write(lu,'(I5,3(F12.6))') tmp_ntotal, xinit
    write(lu,'(I5,3(F12.6))') nstep(1), xinc(1), 0d0, 0d0
    write(lu,'(I5,3(F12.6))') nstep(2), 0d0, xinc(2), 0d0
    write(lu,'(I5,3(F12.6))') nstep(3), 0d0, 0d0, xinc(3)
    do i = 1, nfiles
       do j = 1, m(i)%n
          tmp_count = tmp_count + 1
          if( tmp_count > 99999) then
             exit
          end if
          write(lu,'(I4,F5.1,F11.6,F11.6,F11.6)') m(i)%z(j), 0d0, m(i)%x(:,j)
       end do
       if( tmp_count > 99999) then
          exit
       end if
    end do

  end subroutine write_cube_header

!  subroutine write_cpt_body(anci)
!
!    type(averagenci), intent(in) :: anci
!    integer :: i, fpt, do_print
!    
!    fpt = 3
!    do_print = 1
!    open(fpt, file = anci%cptfile)
!    do i=1,anci%cptcount
!!        do j=1,anci%cptcount
!!            if (j /= i) then 
!!                if ( abs(anci%criticalpt(2,i)-anci%criticalpt(2,j)) < 1.5*anci%xinc(1) &
!!               .and. abs(anci%criticalpt(3,i)-anci%criticalpt(3,j)) < 1.5*anci%xinc(2) &
!!               .and. abs(anci%criticalpt(4,i)-anci%criticalpt(4,j)) < 1.5*anci%xinc(3)) then
!!                   do_print = 0
!!                end if
!!            end if
!!        end do
!        
!        if(do_print .eq. 1) then
!            write(fpt, '(I5, 3F10.4, 2E15.4)') int(anci%criticalpt(1,i)), anci%criticalpt(2,i),  &
!            anci%criticalpt(3,i), anci%criticalpt(4,i), anci%criticalpt(5,i), anci%criticalpt(6,i)
!        end if
!        do_print = 1
!    end do
!    close(fpt)
!
!
!  end subroutine write_cpt_body

  subroutine write_cube_body(lu,n,c)
    
    integer, intent(in) :: lu
    integer, intent(in) :: n(3)
    real*8, intent(in) :: c(0:n(3)-1,0:n(2)-1,0:n(1)-1)

    integer :: i, j

    do i = 0, n(1)-1
       do j = 0, n(2)-1
          write (lu,'(6(1x,e12.5))') (c(k,j,i),k=0,n(3)-1)
       enddo
    enddo
    close(lu)

  end subroutine write_cube_body

  subroutine screengrid(switcher)
    
    logical, intent(out)::switcher
    real*8 :: dist(3)
    real*8 :: distance
    integer:: i
    real*8 :: rthres2

    rthres2 = rthres * rthres
    switcher = .false.


    do i = 1, m(udat0)%n
        dist = x - m(udat0)%x(:,i)
        distance = dot_product(dist, dist)
!write(*,*) x
!write(*,*) m(udat0)%x(:,i)
!write(*,*) dist
!write(*,*) distance, rthres2
        if(distance <= rthres2) then
            switcher = .true.
            exit
        end if
    end do

  end subroutine screengrid
    

end program
